/*
 * Printer Filter for Pentax PocketJet / PocketJet II
 * (C) Copyright 2001, Rildo Pragana <rildo@pragana.net>
 * This is free software, released under the General Public License (GPL)
 * The development of this work was sponsored by Pentax Technologies.
 */

#include "pentaxpj.h"

int lp=0;
FILE *bitmapf;
int bmwidth=0, bmheight=0;
int lpwidth=308, lpheight=3200;
unsigned char buf[1000];
unsigned char cmdbuf[10]={'\033','~'};
char *gscommands[] = { "/usr/bin/gs", "/bin/gs", "/usr/local/bin/gs", NULL };
char *gscmd=NULL;
char res_str[30]="-r300x300";
char *lpdevice=NULL;
int optT=0, optd=175, optR=0, optP=0;
char *opto=NULL;
int tmpwidth=0, tmpheight=0;
int serial_flg=0;

/*
 * locate ghostscript program in a path list
 */
void
find_gs () {
	char **tmp;
	tmp = gscommands; 
	while (*tmp != NULL) {
		if (!access(*tmp,X_OK)) {
			break;
		}
		tmp++;
	}
	gscmd = *tmp;
	if (gscmd == NULL) {
		fprintf(stderr,"*** ghostscript interpreter (gs) not found!\n");
		exit(1);
	}
}

/*
 * low-level commands to the printer
 */
void
cmdreset () {
	write(lp,"\033@",2);
}

void
cmdh1 (int cmd, int parm) {
	sprintf(cmdbuf+2,"%c%c",cmd,parm);
	write(lp,cmdbuf,4);
}

void
cmdh2 (int cmd, int parm) {
	sprintf(cmdbuf+2,"%c%c%c",cmd,parm % 256, parm / 256);
	write(lp,cmdbuf,5);
}

/*
 * read_bitmap -- read buffer from the bitmapf channel
 *      n = 0 --> get a single line at most HDRLINE_SIZE
 *      n > 0 --> read the number of bytes given by n
 * Returns -1 if error found, 0 otherwise
 */
int
read_bitmap (int n) {
	struct pollfd ufds;
	char *s;
	int cnt;
	if (!n) {
		s = buf;
		cnt = HDRLINE_SIZE;
		while (cnt--) {
			ufds.fd = fileno(bitmapf);
			ufds.events = POLLIN;
			if (poll(&ufds,1,PSHDR_TIMEOUT) <= 0)
				return -1;
			*s = fgetc(bitmapf);
			if (*s == '\n') {
				*++s=0;
				break;
			}
			s++;
		}
	}
	else {
		s=buf;
		while (n) { 
			ufds.fd = fileno(bitmapf);
			ufds.events = POLLIN;
			if (poll(&ufds,1,PSDATA_TIMEOUT) <= 0)
				return -1;
			n -= fread(s,1,n,bitmapf);
			s += n;
		}
	}
	return 0;
}

/*
 * find a cluster of at least "n" null bytes 
 * in a string "s" of a given "size"
 */
unsigned char *
find_zeros( unsigned char *s, int size, int n ) {
	register unsigned char *r,*t;
	register int i=0;
	for (r=s ; r < (s+size-n) ; r+=(i+1) ) {
		for (t=r,i=0;i<n;i++) 
			if (*t++) break;
		if (i==n) return r;
	}
	return NULL;
}

/*
 * skip "n" pixel lines
 */
void
skip_lines ( int n ) {
	while (n>255) { 
		/*ESC~J\x01       Flush line buffer and advance to next line*/
		cmdh1('J',255);
		n -= 255;
	}
	if (n) 
		cmdh1('J',n);
}

void
send_serial( int lp, char *buf, int count ) {
	while (count > SERBLK_SIZE) {
		write( lp, buf, SERBLK_SIZE );
		buf += SERBLK_SIZE;
		count -= SERBLK_SIZE;
		usleep(SERBLK_DELAY);
	}
	write( lp, buf, count );
}

/*
 * send bitmap page(s) got from gs pipe to the printer
 */
int
send_bitmap () {
	int i,j;
	int xc=0,yc=0,w,n,wmax,newln=0;
	unsigned char *s,*t;

	while (!feof(bitmapf)) {
		if (read_bitmap(0))
			break;
		if (strncmp(buf,"P4",2)) {
			return 0;
		}
		/* bypass the comment line */
		do {
			if (read_bitmap(0)) {
				fprintf(stderr,"Bitmap file seems truncated.\n");
				exit(1);
			}
		} while (buf[0] == '#');
		/* read the bitmap's dimensions */
		if (sscanf(buf,"%d %d",&bmwidth,&bmheight)<2) {
			fprintf(stderr,"Bitmap file with wrong size fields.\n");
			exit(1);
		}
		bmwidth = (bmwidth+7)/8;
		/*fprintf(stderr,"width = %d, height= %d\n",bmwidth, bmheight);	*/
		wmax = min(lpwidth,bmwidth);

		for (i=0;i<min(lpheight,bmheight);i++) {
			/*fprintf(stderr,"line %d\n",i);
			fflush(stderr);*/
			if (read_bitmap(bmwidth))
				break;
			/* send compressed line to device */
			s = t = buf;
			w = wmax;
			while (w) {
				n = COMP_THRESHOLD; 
				t = find_zeros( s, w, n);
				if ((t > s) || (t==NULL)) {
					if (t==NULL) {
						t = s+w;
					}
					if (yc) {
						skip_lines( yc );
						yc=0;
					}
					/*ESC~$##         Set cursor X  command*/
					cmdh2('$',(s-buf)*8);
					/* send pixels till next zeros cluster */
					newln++;
					/*ESC~*++++       Image Data transfer command.*/
					cmdh2('*',t-s);
					if (serial_flg) {
						send_serial(lp,s,t-s);
					}
					else {
						write(lp,s,t-s);
					}
					w -= (t-s);
					s = t;
				}
				if (w) {
					/* locate next not null pixels cluster */
					while (*(s+n) == 0) {
						if (n == w)
							break;
						n++;
					}
					if (n == wmax) {
						yc++;
						break;
					}
					s += n;
					w -= n;
				}
			}
			if (newln) { 
				skip_lines( 1 );
				newln=0;
			}
		}
		write(lp,"\f",1);
		for (i=min(lpheight,bmheight);i<bmheight;i++) {
			if (read_bitmap(bmwidth))
				break;
		}
	}
	return 0;
}

void
setup_printer () {
    /*ESC~q#      Hex(1)      Print Quality Set.*/
	cmdh1('q',optT);
    /*ESC~d##     Hex(2)      Set print density level(0-255).*/
	cmdh2('d',optd);
    /*ESC~w##      Set Page Width (in bytes)*/
	cmdh2('w',lpwidth);
    /*ESC~h##      Set Page Height (in raster lines)*/
	cmdh2('h',lpheight);
    /*ESC~f#       Set Paper Eject Mode (or Form Feed Mode)*/
	cmdh1('f',1);
}

/*
 * detect any anomaly in ghostscript (bad ps document?)
 */
void
gs_handler (int sig) {
	/*fprintf(stderr,"gs abnormally terminated!\n");
	exit(1);*/
}

void
null_handler (int sig) {
	fprintf(stderr,"ignoring SIGCHLD\n");
}

/* 
 * device specific initialization 
 */
void
device_init( ) {
	struct stat stbuf;
	int c, res,maj,m;
	struct termios oldtio,newtio;

	fstat( lp,&stbuf );
	maj = major(stbuf.st_rdev);
	fprintf(stderr,"lpdevice major: %d, minor: %d\n", 
		maj, minor(stbuf.st_rdev));
	if (maj != 4) {
		return;
	}
	serial_flg++;
	/*tcgetattr(lp,&oldtio); /* save current port settings */
	bzero(&newtio, sizeof(newtio));
	newtio.c_cflag = B57600 | CS8 | CSTOPB | CLOCAL | CREAD;
	newtio.c_iflag = IXON | IGNPAR;
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	newtio.c_cc[VTIME]    = 0; 
	newtio.c_cc[VMIN]     = 1;    
	newtio.c_line         = N_TTY;    

	/*tcflush(lp, TCIFLUSH);*/
	tcsetattr(lp,TCSANOW,&newtio);

	/* setup DTR, RTS */
	ioctl(lp, TIOCMGET, &m);
	m |= TIOCM_DTR;
	m &= ~TIOCM_RTS;
	ioctl(lp, TIOCMSET, &m);
}

void
dump_pbmfile( char *fname ) {
	if (signal(SIGCHLD, null_handler) == SIG_ERR) {
		fprintf(stderr, "cannot set handler for SIGCHLD\n");
		exit(1);
	}
	if (fname==NULL) {
		bitmapf = stdin;
	}
	else {
		bitmapf = fopen( fname,"r" );
	}
	fcntl(fileno(bitmapf),F_SETFL,(long)O_NONBLOCK);
	send_bitmap();
	fclose(bitmapf);
}

void
usage() {
	fprintf(stderr,"\nusage: pentaxpj -[2RPTdwhop] file ...\n");
	fprintf(stderr,"option summary:\n");
	fprintf(stderr,"\t-2         set 200x200 pixel resolution\n");
	fprintf(stderr,"\t-R         reset printer and exit\n");
	fprintf(stderr,"\t-P         print raw pbm file\n");
	fprintf(stderr,"\t-T         text quality (default graphics)\n");
	fprintf(stderr,"\t-d dens    set density\n");
	fprintf(stderr,"\t-W width   set paper width (* 8 pixels units)\n");
	fprintf(stderr,"\t-H height  set paper height in pixels\n");
	fprintf(stderr,"\t-o file    send output to a file\n");
	fprintf(stderr,"\t-p device  send output to a device\n");
	exit(1);
}

/* parse command line arguments */
void
parse_opts( int argc, char **argv ) {
	int c;
	while ((c = getopt(argc,argv,"2RPTd:W:H:o:p:cw:l:i:n:h:")) != -1) {
		switch (c) {
			case '2': {
				strcpy(res_str,"-r200x200");
				/* set defaults for letter paper */
				lpwidth = 204;
				lpheight = 2133;
				break;
			}
			case 'P': {
				optP=1;
				break;
			}
			case 'R': {
				optR=1;
				break;
			}
			case 'T': {
				optT=1;
				break;
			}
			case 'd': {
				sscanf(optarg,"%d",&optd);
				/*optd = (optd/25)*25;*/
				optd = optd*25;
				break;
			}
			case 'W': {
				sscanf(optarg,"%d",&tmpwidth);
				break;
			}
			case 'H': {
				sscanf(optarg,"%d",&tmpheight);
				break;
			}
			case 'o': {
				opto = strdup(optarg);
				break;
			}
			case 'p': {
				lpdevice = strdup(optarg);
				break;
			}
			case '?': {
				usage();
				break;
			}
			default: {
				/* ignore lpd flags */
				break;
			}
		}
	}
}

int
main(int argc, char *argv[]) {
	char **fname;
	int optcnt;
	char *optns[30];
	char optbuf[80];
	FILE *cfgf;
	char *s,*sbp;
	int gspipe[2];
	pid_t gspid;
	int stdin_flg=0;

	/* check defaults saved in our config file */
	if (!access(CFGFN,R_OK)) {
		cfgf = fopen( CFGFN, "r" );
		fgetc(cfgf); /* bypass '#' comment */
		optcnt = 0;
		optns[optcnt++] = argv[0];
		fgets( optbuf, 80, cfgf );
		sbp = optbuf;
		while( (s = strsep( &sbp, " \t\n" )) != NULL) {
			optns[optcnt++] = strdup(s);
		}
		optns[--optcnt] = NULL;
		parse_opts( optcnt, optns );
		optind = 0;
	}
	/* defaults will be overriden by cmd line args */
	parse_opts( argc, argv );
	/*if (*(argv+optind) == NULL)
		usage();*/
	if (tmpwidth) {
		lpwidth = tmpwidth;
	}
	if (tmpheight) {
		lpheight = tmpheight;
	}
	if (opto!=NULL) {
		lp = open(opto,O_WRONLY | O_CREAT);
	}
	else {
		if (lpdevice != NULL) {
			lp = open(lpdevice,O_RDWR | O_NOCTTY);
			/*lp = open(lpdevice,O_WRONLY);*/
		}
		else {
			lp = STDOUT_FILENO;
		}
		device_init();
	}
	cmdreset();
	if (optR) {
		close(lp);
		return 0;	
	}
	find_gs();
	/*fprintf(stderr,"gs found at %s\n",gscmd);*/
	setup_printer(); 
	/*fprintf(stderr,"printer setup\n");*/
	if (*(argv+optind) == NULL) {
		stdin_flg++;
	}
	for (fname=argv+optind;stdin_flg || (*fname != NULL); fname++) {
		if (optP) {
			dump_pbmfile( *fname );
		}
		else {
			pipe(gspipe);
			if (signal(SIGCHLD, gs_handler) == SIG_ERR) {
				fprintf(stderr, "cannot set handler for SIGCHLD\n");
				exit(1);
			}
			gspid = fork();
			if (gspid == (pid_t) 0) {
				int nullfd;
				nullfd=open("/dev/null",O_WRONLY);
				close(gspipe[0]);
				dup2(gspipe[1],STDOUT_FILENO);
				dup2(nullfd,STDERR_FILENO);
				if (stdin_flg) {
					execl( gscmd,gscmd,"-q","-dNOPAUSE",res_str,
						"-sDEVICE=pbmraw","-sOutputFile=-","-", NULL ); 
				}
				else {
					execl( gscmd,gscmd,"-q","-dNOPAUSE",res_str,
						"-sDEVICE=pbmraw","-sOutputFile=-",*fname, NULL ); 
				}
			}
			else {
				close(gspipe[1]);
				bitmapf = fdopen(gspipe[0],"r");
				fcntl(fileno(bitmapf),F_SETFL,(long)O_NONBLOCK);
				send_bitmap();
				fclose(bitmapf);
				if (stdin_flg)
					break;
			}
		}
	}
	close(lp);
	return 0;
}

/* end of pentaxpj.c */

